{$I OPTIONS.INC}

unit gpParser;

interface

uses
  Classes,
  EZDSLBSE,
  EZDSLDBL,
  gppIDT;

type
  TNotifyProc = procedure(unitName: string) of object;
  TNotifyInstProc = procedure(fullName, unitName: string; parse: boolean) of object;

  TUnitList = class;
  TProcList = class;
  TAPIList  = class;
  TProject  = class;
  TProc     = class;

  TDefineList = class
    dlDefines: TStringList;
    constructor Create;
    destructor  Destroy; override;
    procedure   Define(conditional: string);
    procedure   Undefine(conditional: string);
    function    IsDefined(conditional: string): boolean;
    procedure   Assign(conditionals: string);
  end;

  TUnit = class
    unName           : string;
    unQual           : string;
    unUnits          : TUnitList;
    unProcs          : TProcList;
    unAPIs           : TAPIList;
    unParsed         : boolean;
    unExcluded       : boolean;
    unProjectDir     : boolean;
    unAllInst        : boolean;
    unNoneInst       : boolean;
    unUsesOffset     : integer;
    unImplementOffset: integer;
    unStartUses      : integer;
    unEndUses        : integer;
    unFileDate       : integer;
    constructor Create(unitName: string; unitLocation: string = ''; excluded: boolean = false);
    destructor  Destroy; override;
    function    Parse(project: TProject; exclUnits, searchPath, defaultDir, conditionals: string; rescan, parseAssembler: boolean): boolean;
    procedure   CheckInstrumentedProcs;
    function    LocateUnit(unitName: string): TUnit;
    function    LocateProc(procName: string): TProc;
    procedure   Instrument(project: TProject; idt: TIDTable; keepDate: boolean);
    procedure   ConstructNames(idt: TIDTable);
    function    AnyInstrumented: boolean;
    function    AnyChange: boolean;
  end;

  TProc = class
    prName          : string;
    prHeaderLineNum : integer;
    prStartOffset   : integer;
    prStartLineNum  : integer;
    prEndOffset     : integer;
    prEndLineNum    : integer;
    prInstrumented  : boolean;
    prInitial       : boolean;
    prCmtEnterBegin : integer;
    prCmtEnterEnd   : integer;
    prCmtExitBegin  : integer;
    prCmtExitEnd    : integer;
    prPureAsm       : boolean;
    constructor Create(procName: string; offset: integer = 0; lineNum: integer = 0; headerLineNum: integer = 0);
    destructor  Destroy; override;
  end;

  TAPI = class
    apiCommands : string;
    apiBeginOffs: integer;
    apiEndOffs  : integer;
    apiExitBegin: integer;
    apiExitEnd  : integer;
    apiMeta     : boolean;
    constructor Create(apiCmd: string; apiBegin, apiEnd, apiExStart, apiExEnd: integer; apiIsMetaComment: boolean);
  end;

  TUnitList = class(TDList)
    constructor Create; reintroduce;
    procedure   Add(anUnit: TUnit);
  end;

  TGlbUnitList = class(TDList)
    constructor Create; reintroduce;
    function    Locate(unitName: string): TUnit;
    function    LocateCreate(unitName, unitLocation: string; excluded: boolean): TUnit;
  end;

  TProcList = class(TDList)
    constructor Create; reintroduce;
    procedure   Add(var procName: string; pureAsm: boolean; offset, lineNum, headerLineNum: integer);
    procedure   AddEnd(procName: string; offset, linenum: integer);
    procedure   AddInstrumented(procName: string; cmtEnterBegin,cmtEnterEnd,cmtExitBegin,cmtExitEnd: integer);
  end;

  TAPIList = class(TDList)
    constructor Create; reintroduce;
    procedure   AddMeta(apiCmd: string; apiBegin, apiEnd: integer);
    procedure   AddExpanded(apiEnterBegin,apiEnterEnd,apiExitBegin,apiExitEnd: integer);
  end;

  TProject = class
    prName            : string;
    prUnit            : TUnit;
    prUnits           : TGlbUnitList;
    prConditStart     : string;
    prConditStartUses : string;
    prConditStartAPI  : string;
    prConditEnd       : string;
    prConditEndUses   : string;
    prConditEndAPI    : string;
    prAppendUses      : string;
    prCreateUses      : string;
    prProfileEnterProc: string;
    prProfileExitProc : string;
    prProfileEnterAsm : string;
    prProfileExitAsm  : string;
    prProfileAPI      : string;
    prAPIIntro        : string;
    constructor Create(projName: string);
    destructor  Destroy; override;
    procedure   Parse(exclUnits, searchPath, conditionals: string; notify: TNotifyProc; commentType: integer; parseAssembler: boolean);
    procedure   Rescan(exclUnits, searchPath, conditionals: string; notify: TNotifyProc; commentType: integer; ignoreFileDate: boolean; parseAssembler: boolean);
    property    Name: string read prName;
    procedure   GetUnitList(s: TStringList; projectDirOnly, getInstrumented: boolean);
    procedure   GetProcList(unitName: string; s: TStringList; getInstrumented: boolean);
    function    GetUnitPath(unitName: string): string;
    procedure   InstrumentAll(instrument, projectDirOnly: boolean);
    procedure   InstrumentUnit(unitName: string; instrument: boolean);
    procedure   InstrumentProc(unitName, procName: string; instrument: boolean);
    procedure   InstrumentTUnit(anUnit: TUnit; instrument: boolean);
    function    AllInstrumented(projectDirOnly: boolean): boolean;
    function    NoneInstrumented(projectDirOnly: boolean): boolean;
    function    AnyInstrumented(projectDirOnly: boolean): boolean;
    procedure   Instrument(projectDirOnly: boolean; notify: TNotifyInstProc;
      commentType: integer; keepDate: boolean;
      incFileName, conditionals, searchPath: string; parseAssembler: boolean);
    function    GetFirstLine(unitName, procName: string): integer;
    function    AnyChange(projectDirOnly: boolean): boolean;
  private
    procedure   PrepareComments(commentType: integer);
  end;

implementation

uses
  Windows,
  SysUtils,
{$IFDEF DDP}
  GpExcept,
{$ENDIF}
{$IFDEF LogParser}
  GpIFF,
{$ENDIF}
{$IFDEF DebugParser}
  uDbg,
  uDbgIntf,
{$ENDIF}
  GpString,
  mPasLex,
  gppCommon,
  gpFileEdit;

{========================= TDefineList =========================}

  constructor TDefineList.Create;
  begin
    inherited Create;
    dlDefines := TStringList.Create;
    dlDefines.Sorted := true;
  end;

  procedure TDefineList.Define(conditional: string);
  begin
    if not IsDefined(conditional) then dlDefines.Add(UpperCase(conditional));
  end;

  destructor TDefineList.Destroy;
  begin
    dlDefines.Free;
    inherited Destroy;
  end;

  function TDefineList.IsDefined(conditional: string): boolean;
  begin
    Result := (dlDefines.IndexOf(UpperCase(conditional)) >= 0);
  end;

  procedure TDefineList.Undefine(conditional: string);
  var
    idx: integer;
  begin
    idx := dlDefines.IndexOf(UpperCase(conditional));
    if idx >= 0 then dlDefines.Delete(idx);
  end;

  procedure TDefineList.Assign(conditionals: string);
  var
    i: integer;
  begin
    dlDefines.Clear;
    for i := 1 to NumElements(conditionals,';',-1) do
      Define(NthEl(conditionals,i,';',-1));
  end;

{========================= TUnitList =========================}

  function CompareUnit(data1, data2: pointer): integer;
  begin
    Result := StrIComp(PChar(TUnit(data1).unName),PChar(TUnit(data2).unName));
  end; { CompareUnit }

  constructor TUnitList.Create;
  begin
    inherited Create(false);
    Compare := @CompareUnit;
  end; { TUnitList.Create }

  procedure TUnitList.Add(anUnit: TUnit);
  var
    cursor: TListCursor;
  begin
    cursor := SetAfterLast;
    try InsertBefore(cursor,anUnit); except end;
  end; { TUnitList.Add }

{========================= TGlbUnitList =========================}

  procedure DisposeUnit(aData: pointer);
  begin
    TUnit(aData).Free;
  end; { DisposeUnit }

  constructor TGlbUnitList.Create;
  begin
    inherited Create(true);
    Compare := @CompareUnit;
    DisposeData := @DisposeUnit;
  end; { TGlbUnitList.Create }

  function TGlbUnitList.Locate(unitName: string): TUnit;
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
    un := TUnit.Create(unitName);
    try
      if Search(cursor,un) then Locate := TUnit(Examine(cursor))
                           else Locate := nil;
    finally un.Destroy; end;
  end; { TGlbUnitList.Locate }
  
  function TGlbUnitList.LocateCreate(unitName, unitLocation: string; excluded: boolean): TUnit;
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
  {$IFDEF LogParser}GpLogEvent(Format('LocateCreate(%s,%s,%s)',[unitName,unitLocation,IFF(excluded,'true','false')]),FullLogName);{$ENDIF}
    LocateCreate := Locate(unitName);
    if Result = nil then begin
      un := TUnit.Create(unitName,unitLocation,excluded);
      cursor := SetAfterLast;
      InsertBefore(cursor,un);
      cursor := Prev(cursor);
      LocateCreate := TUnit(Examine(cursor));
  {$IFDEF LogParser}GpLogEvent('Created',FullLogName);{$ENDIF}
    end
  {$IFDEF LogParser}else GpLogEvent('Located',FullLogName);{$ENDIF}
    ;
  end; { TGlbUnitList.LocateCreate }

{========================= TProcList =========================}

  procedure DisposeProc(aData: pointer);
  begin
    TProc(aData).Free;
  end; { DisposeProc }

  function CompareProc(data1, data2: pointer): integer;
  begin
    Result := StrIComp(PChar(TProc(data1).prName),PChar(TProc(data2).prName));
  end; { CompareProc }

  constructor TProcList.Create;
  begin
    inherited Create(true);
    Compare := @CompareProc;
    DisposeData := @DisposeProc;
  end; { TProcList.Create }

  procedure TProcList.Add(var procName: string; pureAsm: boolean; offset, lineNum,headerLineNum: integer);
  var
    pr        : TProc;
    pr2       : TProc;
    post      : integer;
    overloaded: boolean;
    cursor    : TListCursor;
  begin
    if Pos('GetThreadPriority',procName) > 0 then
      pureAsm := pureAsm;  
    pr := TProc.Create(procName,offset,linenum,headerLineNum);
    if Search(cursor,pr) then begin
      pr2 := TProc(Examine(cursor));
      pr2.prName := pr2.prName+':1';
      overloaded := true;
      pr.prName := procName+':1';
    end
    else begin
      pr.prName := procName+':1';
      overloaded := Search(cursor,pr);
      if not overloaded then pr.prName := procName;
    end;
    if overloaded then begin // fixup for overloaded procedures
      post:= 1;
      while Search(cursor,pr) do begin
        Inc(post);
        pr.prName := procName+':'+IntToStr(post);
      end;
      procName := pr.prName;
    end;
    with pr do begin
      prCmtEnterBegin := -1;
      prCmtEnterEnd   := -1;
      prCmtExitBegin  := -1;
      prCmtExitEnd    := -1;
      prPureAsm       := pureAsm;
    end;
    cursor := SetAfterLast;
    try InsertBefore(cursor,pr); except pr.Free; end;
  end; { TProcList.Add }

  procedure TProcList.AddEnd(procName: string; offset, linenum: integer);
  var
    pr    : TProc;
    cursor: TListCursor;
  begin
    pr := TProc.Create(procName);
    try
      if Search(cursor,pr) then begin
        with TProc(Examine(cursor)) do begin
          prEndOffset  := offset;
          prEndLineNum := linenum;
          prInitial    := false;
        end;
      end;
    finally pr.Free; end;
  end; { TProcList.AddEnd }

  procedure TProcList.AddInstrumented(procName: string; cmtEnterBegin,cmtEnterEnd,cmtExitBegin,cmtExitEnd: integer);
  var
    pr    : TProc;
    cursor: TListCursor;
  begin
    pr := TProc.Create(procName);
    try
      if Search(cursor,pr) then begin
        with TProc(Examine(cursor)) do begin
          prCmtEnterBegin := cmtEnterBegin;
          prCmtEnterEnd   := cmtEnterEnd;
          prCmtExitBegin  := cmtExitBegin;
          prCmtExitEnd    := cmtExitEnd;
          prInstrumented  := true;
          prInitial       := true;
        end;
      end;
    finally pr.Free; end;
  end; { TProcList.AddInstrumented }

{========================= TProc =========================}

  constructor TProc.Create(procName: string; offset, lineNum, headerLineNum: integer);
  begin
    prName          := procName;
    prHeaderLineNum := headerLineNum;
    prStartOffset   := offset;
    prStartLineNum  := linenum;
    prInstrumented  := false;
  end; { TProc.Create }

  destructor TProc.Destroy;
  begin
  end; { TProc.Destroy }

{========================= TUnit =========================}

  constructor TUnit.Create(unitName: string; unitLocation: string = ''; excluded: boolean = false);
  begin
    unName            := ExtractFileName(unitName);
    if unitLocation = '' then unQual := unitName
                         else unQual := unitLocation;
    unUnits           := TUnitList.Create;
    unProcs           := TProcList.Create;
    unAPIs            := TAPIList.Create;
    unParsed          := false;
    unExcluded        := excluded;
    unAllInst         := false;
    unNoneInst        := true;
    unUsesOffset      := -1;
    unImplementOffset := -1;
    unStartUses       := -1;
    unEndUses         := -1;
    unFileDate        := -1;
  end; { TUnit.Create }

  destructor TUnit.Destroy;
  begin
    unUnits.Free;
    unProcs.Free;
    unAPIs.Free;
  end; { TUnit.Destroy }

  function FindOnPath(uName, pSearchPath, pDefaultDir: string): string;
  var
    i: integer;
    s: string;
    p: string;
  begin
    if Pos('.',UpperCase(uName)) = 0 then uName := uName + '.pas';
    p := MakeBackslash(pDefaultDir);
    if FileExists(p+uName) then FindOnPath := LowerCase(p+uName)
    else if FileExists(uName) then begin
      if ExtractFilePath(uName) = '' then FindOnPath := LowerCase(MakeBackslash(GetCurrentDir)+uName)
                                     else FindOnPath := LowerCase(uName);
    end
    else begin
      for i := 1 to NumElements(pSearchPath,';',-1) do begin
        s := Compress(NthEl(pSearchPath,i,';',-1));
        if Pos('\',s) = 0 then s := p + s;
        s := MakeSmartBackslash(s)+uName;
        if FileExists(s) then begin
          FindOnPath := LowerCase(s);
          Exit;
        end;
      end;
      FindOnPath := '';    
    end;
  end; { FindOnPath }

  function TUnit.Parse(project: TProject; exclUnits, searchPath, defaultDir, conditionals: string;
    rescan, parseAssembler: boolean): boolean;
  var
    isInstrumentationFlag: boolean;
    parser       : TmwPasLex;
    stream       : TMemoryStream;
    state        : (stScan, stParseUses, stScanProc, stScanProc3, stScanProc4,
                    stWaitSemi);
    stateComment : (stWaitEnterBegin, stWaitEnterEnd, stWaitExitBegin,
                    stWaitExitEnd, stWaitExitBegin2, stNone);
    cmtEnterBegin: integer;
    cmtEnterEnd  : integer;
    cmtExitBegin : integer;
    cmtExitEnd   : integer;
    implement    : boolean;
    block        : integer;
    procname     : string;
    proclnum     : integer;
    stk          : string;
    lnumstk      : string;
    procn        : string;
    uun          : string;
    ugpprof      : string;
    unName       : string;
    unLocation   : string;
    tokenID      : TTokenKind;
    tokenData    : string;
    tokenPos     : integer;
    tokenLN      : integer;
    inAsmBlock   : boolean;
    inRecordDef  : boolean;
    prevTokenID  : TTokenKind;
    APIcmd       : string;
    apiStart     : integer;
    apiStartEnd  : integer;
    direct       : string;
    parserStack  : TList;
    skipList     : TList;
    skipping     : boolean;
    conds        : TDefineList;
    incName      : string;

    function IsBlockStartToken(token: TTokenKind): boolean;
    begin
      if inRecordDef and (token = tkCase)
        then Result := false
        else Result := (token = tkBegin) or (token = tkRepeat) or
                       (token = tkCase) or (token = tkTry) or
                       (token = tkAsm) or (token = tkRecord);
      if token = tkAsm then inAsmBlock := true;
      if token = tkRecord then inRecordDef := true;
    end; { IsBlockStartToken }

    function IsBlockEndToken(token: TTokenKind): boolean;
    begin
      IsBlockEndToken := (token = tkEnd) or (token = tkUntil);
      if Result then begin
        inAsmBlock  := false;
        inRecordDef := false;
      end;
    end; { IsBlockEndToken }

    function ExpandLocation(location: string): string;
    begin
      if location = '' then Result := ''
      else begin
        location := Copy(location,2,Length(location)-2);
        if ((Length(location) >= 2) and (location[2] = ':')) or
           (location[1] = '\')
          then Result := location
          else Result := ExpandFileName(MakeBackslash(ExtractFilePath(project.prUnit.unQual))+location);
      end;
    end; { ExpandLocation }

    procedure CreateNewParser(fName: string; searchOnPath: string);
    var
      zero: char;
    begin
      if parser <> nil then begin
        parserStack.Add(parser);
        parserStack.Add(stream);
      end;
      if not FileExists(fName) then begin
        fName := FindOnPath(fName, searchOnPath, defaultDir);
        if not FileExists(fName) then Abort;
      end;
      parser := TmwPasLex.Create;
      stream := TMemoryStream.Create;
      if FileExists(fName) then begin
        try stream.LoadFromFile(fName); except stream.Free; Abort; end;
      end;
      stream.Position := stream.Size;
      zero := #0;
      stream.Write(zero,1);
      parser.Origin := Stream.Memory;
      parser.RunPos := 0;
    end; { CreateNewParser }

    function RemoveLastParser: boolean;
    begin
      parser.Free;
      stream.Free;
      if parserStack.Count > 0 then begin
        parser := TmwPasLex(parserStack[parserStack.Count-2]);
        stream := TMemoryStream(parserStack[parserStack.Count-1]);
        parserStack.Delete(parserStack.Count-1);
        parserStack.Delete(parserStack.Count-1);
        parser.Next;
        Result := true;
      end
      else begin
        parser := nil;
        stream := nil;
        Result := false;
      end;
    end; { RemoveLastParser }

    function ExtractCommentBody(comment: string): string;
    begin
      if comment = '' then Result := ''
      else if comment[1] = '{' then Result := Copy(comment,2,Length(comment)-2)
      else Result := Copy(comment,3,Length(comment)-4);
    end; { ExtractCommentBody }

    function ExtractDirective(comment: string): string;
    begin
      Result := ButFirst(UpperCase(FirstEl(ExtractCommentBody(comment),' ',-1)),1);
      // Fix Delphi stupidity - Delphi parses {$ENDIF.} (where '.' is any
      // non-ident character (alpha, number, underscore)) as {$ENDIF}
      while (Result <> '') and (not (Result[Length(Result)] in ['a'..'z','A'..'Z','0'..'9','_','+','-'])) do
        Delete(Result,Length(Result),1);
    end; { ExtractDirective }

    function ExtractParameter(comment: string; parameter: integer): string;
    begin
      Result := NthEl(ExtractCommentBody(comment),parameter+1,' ',-1);
    end; { ExtractParameter }

    procedure PushSkippingState(skipping: boolean; isIFOPT: boolean);
    begin
      skipList.Add(pointer(skipping));
      skipList.Add(pointer(isIFOPT));
    end; { PushSkippingState }

    function WasSkipping: boolean;
    begin
      if skipList.Count = 0
        then Result := false 
        else Result := boolean(skipList[skipList.Count-2]);
    end; { WasSkipping }

    function InIFOPT: boolean;
    begin
      if skipList.Count = 0
        then Result := false // should not happen, but ...
        else Result := boolean(skipList[skipList.Count-1]);
    end; { TUnit.InIFOPT }

    function PopSkippingState: boolean;
    begin
      if skipList.Count = 0 then Result := true //source damaged - skip the rest
      else begin
        skipList.Delete(skipList.Count-1);
        Result := boolean(skipList[skipList.Count-1]);
        skipList.Delete(skipList.Count-1);
      end;
    end; { PopSkippingState }

    function IsOneOf(key: string; compareWith: array of string): boolean;
    var
      i: integer;
    begin
      Result := false;
      for i := Low(compareWith) to High(compareWith) do begin
        if key = compareWith[i] then begin
          Result := true;
          break;
        end;
      end; //for
    end; { IsOneOf }

  begin
    unParsed := true;
    Parse := true;
    try
      parserStack := TList.Create;
      try
  {$IFDEF LogParser}GpLogEvent(Format('Locating: %s',[unQual]),FullLogName);{$ENDIF}
        if not rescan then begin
          unQual := FindOnPath(unQual, searchPath, defaultDir);
  {$IFDEF LogParser}GpLogEventEx(Format('FindOnPath(%s,%s)=%s',[searchPath,defaultDir,unQual]),FullLogName);{$ENDIF}
          if not FileExists(unQual) then Abort;
          unProjectDir := (self = project.prUnit) or (ExtractFilePath(unQual) = ExtractFilePath(project.prUnit.unQual));
        end
        else begin
          unProcs.Free;
          unProcs := TProcList.Create;
          unUnits.Free;
          unUnits := TUnitList.Create;
        end;
        unAPIs.Free;
        unAPIs := TAPIList.Create;
        parser := nil;
        CreateNewParser(unQual,'');
  {$IFDEF LogParser}GpLogEvent(Format('Parsing: %s',[unQual]),FullLogName);{$ENDIF}
        unFileDate        := FileAge(unQual);
        state             := stScan;
        stateComment      := stNone;
        implement         := false;
        block             := 0;
        procname          := '';
        proclnum          := -1;
        stk               := '';
        lnumstk           := '';
        ugpprof           := UpperCase(profUnitName);
        cmtEnterBegin     := -1;
        cmtEnterEnd       := -1;
        cmtExitBegin      := -1;
        cmtExitEnd        := -1;
        unStartUses       := -1;
        unEndUses         := -1;
        unName            := '';
        unLocation        := '';
        unUsesOffset      := -1;
        unImplementOffset := -1;
        unStartUses       := -1;
        unEndUses         := -1;
        inAsmBlock        := false;
        inRecordDef       := false;
        prevTokenID       := tkNull;
        apiStart          := -1;
        apiStartEnd       := -1;
        skipping          := false;
        skipList := TList.Create;
        try
          conds := TDefineList.Create;
          conds.Assign(conditionals);
          try
            repeat
              while parser.TokenID <> tkNull do begin
                tokenID   := parser.TokenID;
                tokenData := parser.Token;
                tokenPos  := parser.TokenPos;
                tokenLN   := parser.LineNumber;
//if tokenData = '{$DEFINE __breakpoint}' then asm int 3; end;
                {$IFDEF DebugParser}
                  Debugger.LogFmtMsg('Token: %d %s %d %d', [Ord(tokenID),tokenData,tokenPos,tokenLN]);
                {$ENDIF}
                if tokenID = tkCompDirect then begin
                  // Don't process conditional compilation directive if it is
                  // actually an instrumentation flag!
                  with project do
                    isInstrumentationFlag :=
                      IsOneOf(tokenData,[prConditStart,prConditStartUses,prConditStartAPI,
                                         prConditEnd,prConditEndUses,prConditEndAPI]);
                  if not isInstrumentationFlag then begin
                    direct := ExtractDirective(tokenData);
                    if direct = 'IFDEF' then begin //process $IFDEF
                      PushSkippingState(skipping,false);
                      skipping := skipping or (not conds.IsDefined(ExtractParameter(tokenData,1)));
                    end
                    else if direct = 'IFOPT' then begin // process $IFOPT
                      PushSkippingState(skipping,true);
                    end
                    else if direct = 'IFNDEF' then begin //process $IFNDEF
                      PushSkippingState(skipping,false);
                      skipping := skipping or conds.IsDefined(ExtractParameter(tokenData,1));
                    end
                    else if direct = 'ENDIF' then begin //process $ENDIF
                      skipping := PopSkippingState;
                    end
                    else if direct = 'ELSE' then begin //process $ELSE
                      if (not InIFOPT) and (not WasSkipping) 
                        then skipping := not skipping;
                    end;
                  end;
                end;
                if not skipping then begin //we're not in the middle of conditionally removed block
                  if (tokenID = tkPoint) and (prevTokenID = tkEnd) then break; // final end.
                  if tokenID = tkCompDirect then begin
                    // direct := ExtractDirective(tokenData); // Already done
                    if (direct = 'INCLUDE') or (direct = 'I') then begin //process $INCLUDE
                      // process {$I *.INC}
                      incName := ExtractParameter(tokenData,1);
                      if FirstEl(incName,'.',-1) = '*'
                        then incName := FirstEl(ExtractFileName(unQual),'.',-1)+'.'+ButFirstEl(incName,'.',-1);
                      CreateNewParser(incName,searchPath);
                      continue;
                    end
                    else if direct = 'DEFINE' then begin //process $DEFINE
                      conds.Define(ExtractParameter(tokenData,1));
                    end
                    else if direct = 'UNDEF' then begin //process $UNDEF
                      conds.Undefine(ExtractParameter(tokenData,1));
                    end;
                  end;
                  if inAsmBlock and ((prevTokenID = tkAddressOp) or (prevTokenID = tkDoubleAddressOp))
                     and (tokenID <> tkAddressOp) and (tokenID <> tkDoubleAddressOp) then
                    tokenID := tkIdentifier;
                  //fix mwPasParser's feature - these are not reserved words!
                  if (tokenID = tkRead)      or (tokenID = tkWrite)     or
                     (tokenID = tkName)      or (tokenID = tkIndex)     or
                     (tokenID = tkStored)    or (tokenID = tkReadonly)  or
                     (tokenID = tkResident)  or (tokenID = tkNodefault) or
                     (tokenID = tkAutomated) or (tokenID = tkWriteonly)
                    then tokenID := tkIdentifier;
                  if (tokenID = tkBorComment) or (tokenID = tkAnsiComment) or
                     (tokenID = tkCompDirect) then begin
                    if tokenData = project.prConditStartUses then unStartUses := tokenPos
                    else if tokenData = project.prConditEndUses then unEndUses := tokenPos
                    else if ((tokenID = tkBorComment) and (Copy(tokenData,1,1+Length(project.prAPIIntro))='{'+project.prAPIIntro)) or
                            ((tokenID = tkAnsiComment) and (Copy(tokenData,1,2+Length(project.prAPIIntro))='(*'+project.prAPIIntro)) then begin
                      if tokenID = tkBorComment
                        then APIcmd := TrimL(Trim(ButLast(ButFirst(tokenData,1+Length(project.prAPIIntro)),1)))
                        else APIcmd := TrimL(Trim(ButLast(ButFirst(tokenData,2+Length(project.prAPIIntro)),2)));
                      if parserStack.Count = 0 then
                        unAPIs.AddMeta(APIcmd,tokenPos,tokenPos+Length(tokenData)-1);
                    end
                    else if tokenData = project.prConditStartAPI then begin
                      apiStart := tokenPos;
                      apiStartEnd := tokenPos+Length(tokenData)-1;
                    end
                    else if tokenData = project.prConditEndAPI then
                      if parserStack.Count = 0 then
                        unAPIs.AddExpanded(apiStart,apiStartEnd,tokenPos,tokenPos+Length(tokenData)-1);
                  end;
                  if state = stWaitSemi then begin
                    if tokenID = tkSemicolon then begin
                      unImplementOffset := tokenPos+1;
                      state := stScan;
                    end;
                  end
                  else if state = stParseUses then begin
                    if (tokenID = tkSemicolon) or (tokenID = tkComma) then begin
                      if tokenID = tkSemicolon then state := stScan;
                      if unName <> '' then begin
                        uun := UpperCase(unName);
                        unUnits.Add(project.prUnits.LocateCreate(unName,ExpandLocation(unLocation),
                                    (Pos(#13#10+uun+#13#10,exclUnits) <> 0) or (uun = ugpprof)));
                      end;
                      unName := '';
                      unLocation := '';
                    end
                    else if tokenID = tkIdentifier then begin
                      unName := tokenData;
                    end
                    else if tokenID = tkString then begin
                      unLocation := tokenData;
                    end;
                  end
                  else if state = stScanProc then begin
                    if (tokenID = tkIdentifier) or (tokenID = tkPoint) or
                       (tokenID = tkRegister) then begin
                      procname := procname + tokenData;
                      proclnum := tokenLN;
                    end
                    else if ((tokenID = tkSpace) and (procname <> '')) or (tokenID = tkSemicolon) or
                            (tokenID = tkRoundOpen) or (tokenID = tkColon) then begin
                      state := stScanProc3;
                    end;
                  end
                  else if state = stScanProc3 then begin
                    if ((tokenID = tkProcedure) or (tokenID = tkFunction) or
                        (tokenID = tkConstructor) or (tokenID = tkDestructor)) and implement then begin
                      state := stScanProc;
                      block := 0;
                      if procname <> '' then begin
                        stk := stk + '/' + procname;
                        lnumstk := lnumstk + '/' + IntToStr(proclnum);
                      end;
                      procname := '';
                      proclnum := -1;
                    end
                    else if (tokenID = tkForward) or (tokenID = tkExternal) then begin
                      procname := '';
                      proclnum := -1;
                    end
                    else begin
                      if IsBlockStartToken(tokenID) then Inc(block)
                      else if IsBlockEndToken(tokenID) then Dec(block);
                      if block < 0 then begin
                        state    := stScan;
                        stk      := '';
                        lnumstk  := '';
                        procname := '';
                        proclnum := -1;
                      end
                      else if (block > 0) and (not inRecordDef) then begin
                        if stk <> '' then procn := ButFirst(stk,1)+'/'+procname
                                     else procn := procname;
                        if parserStack.Count = 0 then
                          if (tokenID <> tkAsm) or parseAssembler then
                            unProcs.Add(procn,(tokenID = tkAsm),tokenPos,tokenLN,proclnum);
                        state := stScanProc4;
                        stateComment := stWaitEnterBegin;
                      end;
                    end;
                  end
                  else if state = stScanProc4 then begin
                    if IsBlockStartToken(tokenID) then Inc(block)
                    else if IsBlockEndToken(tokenID) then Dec(block);
                    if (tokenID = tkBorComment) or (tokenID = tkCompDirect) then begin
                      if tokenData = project.prConditStart then begin
                        if stateComment = stWaitEnterBegin then begin
                          cmtEnterBegin := tokenPos;
                          stateComment := stWaitEnterEnd;
                        end
                        else if (stateComment = stWaitExitBegin) or (stateComment = stWaitExitBegin2) then begin
                          cmtExitBegin := tokenPos;
                          stateComment := stWaitExitEnd;
                        end;
                      end
                      else if tokenData = project.prConditEnd then begin
                        if stateComment = stWaitEnterEnd then begin
                          cmtEnterEnd := tokenPos;
                          stateComment := stWaitExitBegin;
                        end
                        else if stateComment = stWaitExitEnd then begin
                          cmtExitEnd := tokenPos;
                          stateComment := stWaitExitBegin2;
                        end;
                      end;
                    end;
                    if block = 0 then begin
                      if parserStack.Count = 0 then begin
                        unProcs.AddEnd(procn,tokenPos,tokenLN);
                        if stateComment = stWaitExitBegin2 then begin
                          unProcs.AddInstrumented(procn,cmtEnterBegin,cmtEnterEnd,cmtExitBegin,cmtExitEnd);
                          CheckInstrumentedProcs;
                        end;
                      end;
                      stateComment := stNone;
                      if stk = '' then begin
                        procname := '';
                        proclnum := -1;
                        state    := stScan;
                      end
                      else begin
                        procname := LastEl(stk,'/',-1);
                        proclnum := StrToInt(LastEl(lnumstk,'/',-1));
                        stk      := ButLastEl(stk,'/',-1);
                        lnumstk  := ButLastEl(lnumstk,'/',-1);
                        state    := stScanProc3;
                      end;
                    end;
                  end
                  else if (tokenID = tkUses) or
                          ((tokenID = tkIdentifier) and (UpperCase(tokenData) = 'CONTAINS'){not recognized by the parser}) then begin
                    state := stParseUses;
                    if implement then unUsesOffset := tokenPos;
                  end
                  else if tokenID = tkImplementation then begin
                    implement := true;
                    unImplementOffset := tokenPos;
                  end
                  else if tokenID = tkProgram then begin
                    implement := true;
                    state := stWaitSemi;
                  end
                  else if ((tokenID = tkProcedure) or (tokenID = tkFunction) or
                           (tokenID = tkConstructor) or (tokenID = tkDestructor)) and implement then begin
                    state := stScanProc;
                    block := 0;
                    if procname <> '' then begin
                      stk := stk + '/' + procname;
                      lnumstk := lnumstk + '/' + IntToStr(proclnum);
                    end;
                    procname := '';
                    proclnum := -1;
                  end;
                end; // if not skipping
                prevTokenID := tokenID;
                parser.Next;
              end; // while
            until not RemoveLastParser;
          finally conds.Free; end;
        finally skipList.Free; end;
      finally parserStack.Free; end;
    except Parse := false; end;
  end; { TUnit.Parse }

  procedure TUnit.CheckInstrumentedProcs;
  var
    cursor: TListCursor;
  begin
    unAllInst := true;
    unNoneInst := true;
    with unProcs do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        if not TProc(Examine(cursor)).prInstrumented then unAllInst := false
                                                     else unNoneInst := false;
        cursor := Next(cursor);
      end;
    end;
  end; { TUnit.CheckInstrumentedProcs }

  function TUnit.LocateUnit(unitName: string): TUnit;
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
    un := TUnit.Create(unitName);
    try
      if unUnits.Search(cursor,un) then LocateUnit := TUnit(unUnits.Examine(cursor))
                                   else LocateUnit := nil;
    finally un.Destroy; end; 
  end; { TUnit.LocateUnit }

  function TUnit.LocateProc(procName: string): TProc;
  var
    pr    : TProc;
    cursor: TListCursor;
  begin
    pr := TProc.Create(procName);
    try
      if unProcs.Search(cursor,pr) then LocateProc := TProc(unProcs.Examine(cursor))
                                   else LocateProc := nil;
    finally pr.Destroy; end;
  end; { TUnit.LocateProc }

  procedure TUnit.ConstructNames(idt: TIDTable);
  var
    pr    : TProc;
    cursor: TListCursor;
  begin
    cursor := unProcs.Next(unProcs.SetBeforeFirst);
    while not unProcs.IsAfterLast(cursor) do begin
      pr := TProc(unProcs.Examine(cursor));
      if pr.prInstrumented then idt.ConstructName(unName,unQual,pr.prName,pr.prHeaderLineNum);
      cursor := unProcs.Next(cursor);
    end;
  end; { TUnit.ConstructNames }

  procedure TUnit.Instrument(project: TProject; idt: TIDTable; keepDate: boolean);
  var
    pr      : TProc;
    any     : boolean;
    haveUses: boolean;
    haveInst: boolean;
    ed      : TFileEdit;
    name    : integer;
    justName: string;
    api     : TAPI;
    cursor  : TListCursor;
  begin { TUnit.Instrument }
    if unImplementOffset = -1 then raise Exception.Create('No implementation part defined in unit '+unName+'!');
    justName := ButLastEl(unQual,'.',Ord(-1));
    DeleteFile(justName+'.bk2');
    RenameFile(justName+'.bk1',justName+'.bk2');
    CopyFile(PChar(unQual),PChar(justName+'.bk1'),false);
    ed := TFileEdit.Create(unQual);
    try
      any := AnyInstrumented;
      haveUses := (unStartUses >= 0) and (unEndUses > unStartUses);
      if haveUses and (not any) then ed.Remove(unStartUses,unEndUses+Length(project.prConditEndUses)-1)
      else if (not haveUses) and any then begin
        if LocateUnit(profUnitName) = nil then begin
          if unUsesOffset <> -1
            then ed.Insert(unUsesOffset+Length('uses'),project.prAppendUses)
            else ed.Insert(unImplementOffset+Length('implementation'),project.prCreateUses);
        end;
      end;

      cursor := unAPIs.Next(unAPIs.SetBeforeFirst);
      while not unAPIs.IsAfterLast(cursor) do begin
        api := TAPI(unAPIs.Examine(cursor));
        if any then begin
          if api.apiMeta then begin
            ed.Remove(api.apiBeginOffs,api.apiEndOffs);
            ed.Insert(api.apiBeginOffs,Format(project.prProfileAPI,[api.apiCommands]));
          end
        end
        else begin
          if not api.apiMeta then begin
            ed.Remove(api.apiBeginOffs,api.apiEndOffs);
            ed.Insert(api.apiBeginOffs,'{'+project.prAPIIntro);
            ed.Remove(api.apiExitBegin,api.apiExitEnd);
            ed.Insert(api.apiExitBegin,'}');
          end;
        end;
        cursor := unAPIs.Next(cursor);
      end;

      cursor := unProcs.Next(unProcs.SetBeforeFirst);
      while not unProcs.IsAfterLast(cursor) do begin
        pr := TProc(unProcs.Examine(cursor));
        haveInst := (pr.prCmtEnterBegin >= 0);
        if not pr.prInstrumented then begin
          if haveInst then begin // remove instrumentation
            ed.Remove(pr.prCmtEnterBegin,pr.prCmtEnterEnd+Length(project.prConditEnd)-1);
            ed.Remove(pr.prCmtExitBegin,pr.prCmtExitEnd+Length(project.prConditEnd)-1);
          end;
        end
        else begin
          name := idt.ConstructName(unName,unQual,pr.prName,pr.prHeaderLineNum);
          if haveInst then ed.Remove(pr.prCmtEnterBegin,pr.prCmtEnterEnd+Length(project.prConditEnd)-1);
          if pr.prPureAsm then ed.Insert(pr.prStartOffset+Length('asm'),Format(project.prProfileEnterAsm,[name]))
                          else ed.Insert(pr.prStartOffset+Length('begin'),Format(project.prProfileEnterProc,[name]));
          if haveInst then ed.Remove(pr.prCmtExitBegin,pr.prCmtExitEnd+Length(project.prConditEnd)-1);
          if pr.prPureAsm then ed.Insert(pr.prEndOffset,Format(project.prProfileExitAsm,[name]))
                          else ed.Insert(pr.prEndOffset,Format(project.prProfileExitProc,[name]));
        end;
        cursor := unProcs.Next(cursor);
      end;
      ed.Execute(keepDate);
    finally ed.Free; end;
  end; { TUnit.Instrument }

  function TUnit.AnyInstrumented: boolean;
  var
    cursor: TListCursor;
  begin
    Result := false;
    with unProcs do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        if TProc(Examine(cursor)).prInstrumented then begin
          Result := true;
          Exit;
        end;
        cursor := Next(cursor);
      end; //while
    end; //with
  end; { TUnit.AnyInstrumented }

  function TUnit.AnyChange: boolean;
  var
    pr    : TProc;
    cursor: TListCursor;
  begin
    Result := false;
    with unProcs do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        pr := TProc(Examine(cursor));
        if pr.prInstrumented <> pr.prInitial then begin
          Result := true;
          Exit;
        end;
        cursor := Next(cursor);
      end; //while
    end; //with
  end; { TUnit.AnyChange }

{========================= TProject =========================}

  constructor TProject.Create(projName: string);
  begin
    prUnits:= TGlbUnitList.Create;
    prName := projName;
    prUnit := nil;
  end; { TProject.Create }

  destructor TProject.Destroy;
  begin
    prUnits.Free;
  end; { TProject.Destroy }

  procedure TProject.Parse(exclUnits, searchPath, conditionals: string;
    notify: TNotifyProc; commentType: integer; parseAssembler: boolean);
  var
    un    : TUnit;
    u1    : TUnit;
    cursor: TListCursor;
    curDir: string;
  begin
    PrepareComments(commentType);
    if Last(exclUnits,2) <> #13#10 then exclUnits := exclUnits + #13#10;
    if First(exclUnits,2) <> #13#10 then exclUnits := #13#10 + exclUnits;
    prUnits.Empty;
    prUnit := prUnits.LocateCreate(prName,'',false);
    prUnit.unProjectDir := true;
    GetDir(0,curDir);
    ChDir(ExtractFilePath(prUnit.unQual));
    try
      un := prUnit;
      repeat
        if @Notify <> nil then Notify(un.unName);
        un.Parse(self,exclUnits,searchPath,ExtractFilePath(prName),conditionals,false,parseAssembler);
        cursor := prUnits.Next(prUnits.SetBeforeFirst);
        un := nil;
        while not prUnits.IsAfterLast(cursor) do begin
          u1 := TUnit(prUnits.Examine(cursor));
          if not (u1.unParsed or u1.unExcluded) then begin
            un := u1;
            break;
          end;
          cursor := prUnits.Next(cursor);
        end;
      until (un = nil);
    finally ChDir(curDir); end;
  end; { TProject.Parse }

  procedure TProject.GetUnitList(s: TStringList; projectDirOnly, getInstrumented: boolean);
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
    s.Clear;
    with prUnits do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        un := TUnit(Examine(cursor));
        if (not un.unExcluded) and (un.unProcs.Count > 0) and
           ((not projectDirOnly) or un.unProjectDir) then
           if getInstrumented then s.Add(un.unName+IntToStr(Ord(un.unAllInst))+IntToStr(Ord(un.unNoneInst)))
                              else s.Add(un.unName);
        cursor := Next(cursor);
      end;
    end;
  end; { TProject.GetUnitList }

  procedure TProject.GetProcList(unitName: string; s: TStringList; getInstrumented: boolean);
  var
    un    : TUnit;
    pr    : TProc;
    cursor: TListCursor;
  begin
    s.Clear;
    un := prUnits.Locate(unitName);
    if un <> nil then begin
      with un.unProcs do begin
        cursor := Next(SetBeforeFirst);
        while not IsAfterLast(cursor) do begin
          pr :=TProc(Examine(cursor));
          if getInstrumented then s.Add(pr.prName+IntToStr(Ord(pr.prInstrumented)))
                             else s.Add(TProc(Examine(cursor)).prName);
          cursor := Next(cursor);
        end;
      end;
    end;
  end; { TProject.GetProcList }

  procedure TProject.InstrumentTUnit(anUnit: TUnit; instrument: boolean);
  var
    cursor: TListCursor;
  begin
    anUnit.unAllInst := instrument;
    anUnit.unNoneInst := not instrument;
    with anUnit.unProcs do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        TProc(Examine(cursor)).prInstrumented := instrument;
        cursor := Next(cursor);
      end;
    end;
  end; { TProject.InstrumentTUnit }

  procedure TProject.InstrumentAll(instrument, projectDirOnly: boolean);
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
    with prUnits do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        un := TUnit(Examine(cursor));
        if (not un.unExcluded) and (un.unProcs.Count > 0) then
          if (not projectDirOnly) or un.unProjectDir then InstrumentTUnit(un,instrument);
        cursor := Next(cursor);
      end;
    end;
  end; { TProject.InstrumentAll }

  procedure TProject.InstrumentUnit(unitName: string; instrument: boolean);
  var
    un: TUnit;
  begin
    un := prUnits.Locate(unitName);
    if un <> nil then InstrumentTUnit(un,instrument);
  end; { TProject.InstrumentUnit }

  procedure TProject.InstrumentProc(unitName, procName: string; instrument: boolean);
  var
    un: TUnit;
    pr: TProc;
  begin
    un := prUnits.Locate(unitName);
    if un = nil then raise Exception.Create('Trying to instrument unexistent unit!')
    else begin
      pr := un.LocateProc(procName);
      if pr = nil then raise Exception.Create('Trying to instrument unexistend procedure!')
      else begin
        pr.prInstrumented := instrument;
        un.CheckInstrumentedProcs;
      end;
    end;
  end; { TProject.InstrumentProc }

  function TProject.AllInstrumented(projectDirOnly: boolean): boolean;
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
    Result := false;
    with prUnits do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        un := TUnit(Examine(cursor));
        if (not un.unExcluded) and (un.unProcs.Count > 0) and
           (un.unProjectDir or (not projectDirOnly)) then
          if not un.unAllInst then Exit;
        cursor := Next(cursor);
      end;
    end;
    Result := true;
  end; { TProject.AllInstrumented }

  procedure TProject.Instrument(projectDirOnly: boolean;
    notify: TNotifyInstProc; commentType: integer; keepDate: boolean;
    incFileName, conditionals, searchPath: string; parseAssembler: boolean);
  var
    curDir : string;
    un     : TUnit;
    idt    : TIDTable;
    rescan : TList;
    i      : integer;
    unAny  : boolean;
    anyInst: boolean;
    cursor : TListCursor;
  begin
    PrepareComments(commentType);
    rescan := TList.Create;
    try
      idt := TIDTable.Create;
      try
        GetDir(0,curDir);
        ChDir(ExtractFilePath(prUnit.unQual));
        try
          with prUnits do begin
            anyInst := false;
            cursor := Next(SetBeforeFirst);
            while not IsAfterLast(cursor) do begin
              un := TUnit(Examine(cursor));
              if (not un.unExcluded) and (un.unProcs.Count > 0) then begin
                if @Notify <> nil then Notify(un.unQual,un.unName,false);
                unAny := un.AnyInstrumented;
                if unAny then anyInst := true;
                if un.AnyChange or unAny then begin
                  un.Instrument(self,idt,keepDate);
                  rescan.Add(un);
                end
                else un.ConstructNames(idt);
              end;
              cursor := Next(cursor);
            end;
          end;
          idt.Dump(incFileName);
        finally ChDir(curDir); end;
      finally idt.Free; end;
      if not anyInst then begin
        DeleteFile(incFileName);
        DeleteFile(ReplaceExt(incFileName,'gpd'));
      end;
      for i := 0 to rescan.Count-1 do begin
        if @Notify <> nil then Notify(TUnit(rescan[i]).unQual,TUnit(rescan[i]).unName,true);
        TUnit(rescan[i]).Parse(self,'',searchPath,ExtractFilePath(prName),conditionals,true,parseAssembler);
      end;
    finally rescan.Free; end;
  end; { TProject.Instrument }

  function TProject.NoneInstrumented(projectDirOnly: boolean): boolean;
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
    Result := false;
    with prUnits do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        un := TUnit(Examine(cursor));
        if (not un.unExcluded) and (un.unProcs.Count > 0) and
           (un.unProjectDir or (not projectDirOnly)) then
          if not un.unNoneInst then Exit;
        cursor := Next(cursor);
      end;
    end;
    Result := true;
  end; { TProject.NoneInstrumented }

  function TProject.GetUnitPath(unitName: string): string;
  var
    un: TUnit;
  begin
    un := prUnits.Locate(unitName);
    if un = nil then raise Exception.Create('Trying to get name of unexistent unit!')
                else Result := un.unQual;
  end; { TProject.GetUnitPath }

  procedure TProject.PrepareComments(commentType: integer);
  begin
    case commentType of
      0: begin
        prConditStart     := '{>>GpProfile}';
        prConditStartUses := '{>>GpProfile U}';
        prConditStartAPI  := '{>>GpProfile API}';
        prConditEnd       := '{GpProfile>>}';
        prConditEndUses   := '{GpProfile U>>}';
        prConditEndAPI    := '{GpProfile API>>}';
      end;
      1: begin
        prConditStart     := '{$IFDEF GpProfile}';
        prConditStartUses := '{$IFDEF GpProfile U}';
        prConditStartAPI  := '{$IFDEF GpProfile API}';
        prConditEnd       := '{$ENDIF GpProfile}';
        prConditEndUses   := '{$ENDIF GpProfile U}';
        prConditEndAPI    := '{$ENDIF GpProfile API}';
      end;
    end;
    prAppendUses      := prConditStartUses + ' ' + profUnitName + ', ' + prConditEndUses;
    prCreateUses      := prConditStartUses + ' uses ' + profUnitName + '; ' + prConditEndUses;
    prProfileEnterProc:= prConditStart + ' ' + 'ProfilerEnterProc(%d); try ' + prConditEnd;
    prProfileExitProc := prConditStart + ' finally ProfilerExitProc(%d); end; ' + prConditEnd;
    prProfileEnterAsm := prConditStart + ' pushad; mov eax, %d; call ProfilerEnterProc; popad ' + prConditEnd;
    prProfileExitAsm  := prConditStart + ' push eax; mov eax, %d; call ProfilerExitProc; pop eax ' + prConditEnd;
    prProfileAPI      := prConditStartAPI + '%s' + prConditEndAPI;
    prAPIIntro        := 'GPP:';
  end; { TProject.PrepareComments }

  function TProject.GetFirstLine(unitName, procName: string): integer;
  var
    un: TUnit;
    pr: TProc;
  begin
    un := prUnits.Locate(unitName);
    if un = nil then Result := -1
    else begin
      pr := un.LocateProc(procName);
      if pr = nil then Result := -1
                  else Result := pr.prHeaderLineNum;
    end;
  end; { TProject.GetFirstLine }

  function TProject.AnyInstrumented(projectDirOnly: boolean): boolean;
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
    Result := true;
    with prUnits do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        un := TUnit(Examine(cursor));
        if (not un.unExcluded) and (un.unProcs.Count > 0) and
           (un.unProjectDir or (not projectDirOnly)) then
          if un.AnyInstrumented then Exit;
        cursor := Next(cursor);
      end;
    end;
    Result := false;
  end; { TProject.AnyInstrumented }

  procedure TProject.Rescan(exclUnits, searchPath, conditionals: string;
    notify: TNotifyProc; commentType: integer; ignoreFileDate: boolean;
    parseAssembler: boolean);
  var
    un    : TUnit;
    cursor: TListCursor;
    curDir: string;
  begin
    PrepareComments(commentType);
    GetDir(0,curDir);
    ChDir(ExtractFilePath(prUnit.unQual));
    try
      if Last(exclUnits,2) <> #13#10 then exclUnits := exclUnits + #13#10;
      with prUnits do begin
        cursor := Next(SetBeforeFirst);
        while not IsAfterLast(cursor) do begin
          un := TUnit(Examine(cursor));
          if (not un.unExcluded) and (un.unProcs.Count > 0) and
             (ignoreFileDate or (un.unFileDate <> FileAge(un.unQual))) then
            un.Parse(self,exclUnits,searchPath,ExtractFilePath(prName),conditionals,true,parseAssembler);
          cursor := Next(cursor);
        end;
      end;
    finally ChDir(curDir); end;
  end; { TProject.Rescan }

  function TProject.AnyChange(projectDirOnly: boolean): boolean;
  var
    un    : TUnit;
    cursor: TListCursor;
  begin
    Result := true;
    with prUnits do begin
      cursor := Next(SetBeforeFirst);
      while not IsAfterLast(cursor) do begin
        un := TUnit(Examine(cursor));
        if (not un.unExcluded) and (un.unProcs.Count > 0) and
           (un.unProjectDir or (not projectDirOnly)) then
          if un.unFileDate <> FileAge(un.unQual) then Exit;
        cursor := Next(cursor);
      end;
    end;
    Result := false;
  end;

{ TAPI }

  constructor TAPI.Create(apiCmd: string; apiBegin, apiEnd, apiExStart, apiExEnd: integer; apiIsMetaComment: boolean);
  begin
    inherited Create;
    apiCommands  := apiCmd;
    apiBeginOffs := apiBegin;
    apiEndOffs   := apiEnd;
    apiExitBegin := apiExStart;
    apiExitEnd   := apiExEnd;
    apiMeta      := apiIsMetaComment;
  end;

{ TAPIList }

  procedure DisposeAPI(api: TAPI);
  begin
    api.Free;
  end; { DisposeAPI }

  procedure TAPIList.AddExpanded(apiEnterBegin, apiEnterEnd, apiExitBegin,
    apiExitEnd: integer);
  var
    api: TAPI;
  begin
    api := TAPI.Create('', apiEnterBegin, apiEnterEnd, apiExitBegin, apiExitEnd, false);
    InsertBefore(SetAfterLast,api);
  end; { TAPIList.AddExpanded }

  procedure TAPIList.AddMeta(apiCmd: string; apiBegin, apiEnd: integer);
  var
    api: TAPI;
  begin
    api := TAPI.Create(apiCmd, apiBegin, apiEnd, -1, -1, true);
    InsertBefore(SetAfterLast,api);
  end; { TAPIList.AddMeta }

constructor TAPIList.Create;
  begin
    inherited Create(true);
    DisposeData := @DisposeAPI;
  end; { TAPIList.Create }

end.

